@@ -110,9 +110,7 @@ CORBA_Object_release_cb (ORBit_RootObject robj)
 	CORBA_Object obj = (CORBA_Object) robj;
 	CORBA_ORB    orb = obj->orb;
 
-	g_assert (orb != NULL);
-
-	if (obj->profile_list) {
+	if (orb && obj->profile_list) {
 		LINK_MUTEX_LOCK (orb->lock);
 		g_hash_table_remove (orb->objrefs, obj);
 		LINK_MUTEX_UNLOCK (orb->lock);
@@ -211,6 +209,12 @@ ORBit_objref_find (CORBA_ORB   orb,
 	return retval;
 }
 
+static gboolean
+is_localhost (const char *host)
+{
+	return (host && !strcmp (link_get_local_hostname (), host)) ? TRUE : FALSE;
+}
+
 /**
  * ORBit_objref_get_proxy:
  * @obj: the local object
@@ -307,9 +311,15 @@ ORBit_object_get_connection (CORBA_Object obj)
 	gboolean is_ssl = FALSE;
 	GIOPVersion iiop_version = GIOP_1_2;
 	GIOPConnection *cnx = NULL;
+	gboolean unix_socket_enabled = FALSE;
+	gboolean ipv4_ipv6_enabled = FALSE;
+	gboolean unix_socket_failed = FALSE;
 
 	OBJECT_LOCK (obj);
 
+	unix_socket_enabled = ORBit_proto_use ("UNIX");
+	ipv4_ipv6_enabled = (ORBit_proto_use ("IPv4") || ORBit_proto_use ("IPv6"));
+
 	if (obj->connection) {
 		if (ORBit_try_connection_T (obj)) {
 			cnx = obj->connection;
@@ -338,12 +348,33 @@ ORBit_object_get_connection (CORBA_Object obj)
 		if (IOP_profile_get_info (obj, pinfo, &iiop_version, &proto,
 					  &host, &service, &is_ssl, tbuf)) {
 
+			if (unix_socket_failed && ipv4_ipv6_enabled && is_localhost (host))
+				continue;
+
 			obj->connection = giop_connection_initiate (
 				obj->orb, proto, host, service,
 				is_ssl ? LINK_CONNECTION_SSL : 0, iiop_version);
 
+			if (!obj->connection && unix_socket_enabled && ipv4_ipv6_enabled) {
+				if (!strcmp (proto,"UNIX")) 
+					unix_socket_failed = TRUE;
+			}
+	
 			if (obj->connection && ORBit_try_connection_T (obj)) {
-				obj->object_key = objkey;
+				if (!IOP_ObjectKey_equal (obj->object_key, objkey)) {
+					/* We need to remove and re-add the object from the
+					 * objrefs hash table since changing the object key will
+					 * change the object's hash value.
+					 */
+					LINK_MUTEX_LOCK (obj->orb->lock);
+					g_hash_table_remove (obj->orb->objrefs, obj);
+					obj->object_key = objkey;
+					g_hash_table_insert (obj->orb->objrefs, obj, obj);
+					LINK_MUTEX_UNLOCK (obj->orb->lock);
+				} else {
+					obj->object_key = objkey;
+				}
+
 				obj->connection->orb_data = obj->orb;
 
 				dprintf (OBJECTS, "Initiated a connection to '%s' '%s' '%s'\n",
@@ -366,6 +397,7 @@ ORBit_handle_location_forward (GIOPRecvBuffer *buf,
 {
 	GIOPConnection *retval = NULL;
 	GSList         *profiles = NULL;
+	GIOPConnection *old_connection;
 
 	if (ORBit_demarshal_IOR (obj->orb, buf, NULL, &profiles))
 		goto out;
@@ -373,8 +405,15 @@ ORBit_handle_location_forward (GIOPRecvBuffer *buf,
 	OBJECT_LOCK (obj);
 	IOP_delete_profiles (obj->orb, &obj->forward_locations);
 	obj->forward_locations = profiles;
+
+	/* We need to clear the connection because 
+	   forwarding results in a different connection */
+	old_connection = obj->connection;
+	obj->connection = NULL;
 	OBJECT_UNLOCK (obj);
 
+	giop_connection_unref (old_connection);
+
 	retval = ORBit_object_get_connection (obj);
 
  out:
